from typing import Callable, Any

import pika
from marshmallow import Schema
from pycapo import CapoConfig
from workspaces.json import WorkflowEventSchema


def make_amqp_connection(profile: str) -> pika.BlockingConnection:
    """
    Initialize connection to AMQP server given a CAPO profile
    :param profile: Capo profile to use
    :return: Established connection
    """
    capo_cfg = CapoConfig(profile=profile)

    hostname = capo_cfg.getstring('edu.nrao.archive.configuration.AmqpServer.hostname')
    username = capo_cfg.getstring('edu.nrao.archive.configuration.AmqpServer.username')
    password = capo_cfg.getstring('edu.nrao.archive.configuration.AmqpServer.password')

    return pika.BlockingConnection(
        pika.ConnectionParameters(
            host=hostname,
            credentials=pika.PlainCredentials(
                username=username,
                password=password
            )
        )
    )


def configure_amqp_channel(
        connection: pika.BlockingConnection,
        exchange: str,
        exchange_type: str
) -> pika.adapters.blocking_connection.BlockingChannel:
    """
    Takes in a connection and exchange declaration parameters and creates a channel to the given
    exchange, configured with the given parameters
    :param connection: RabbitMQ BlockingConnection
    :param exchange: Exchange name
    :param exchange_type: Exchange type
    :return: Channel for given connection
    """
    channel = connection.channel()
    channel.exchange_declare(exchange=exchange, exchange_type=exchange_type, durable=True)
    return channel


def emit_amqp_message(
        channel: pika.adapters.blocking_connection.BlockingChannel,
        exchange: str,
        routing_key: str,
        msg: bytes
):
    """
    Emit AMQP message to a given exchange through a given channel
    :param channel: AMQP connection channel
    :param exchange: AMQP exchange name
    :param routing_key: Routing key
    :param msg: Message to be emitted
    """
    channel.basic_publish(exchange=exchange, routing_key=routing_key, body=msg)


def receive_amqp_message(
        channel: pika.adapters.blocking_connection.BlockingChannel,
        exchange: str,
        routing_key: str,
        callback: Callable,
        auto_ack: bool = False
):
    """
    Establishes queue and binds it to a given channel and consumes messages matching the
    routing key from given exchange
    :param channel: AMQP connection channel
    :param exchange: AMQP exchange name
    :param routing_key: Routing key
    :param callback: Callback function for when a message is consumed
    :param auto_ack: If true, consumer automatically acknowledges when it consumes a message
    """
    queue_result = channel.queue_declare(queue='', exclusive=True)
    queue_name = queue_result.method.queue
    channel.queue_bind(queue=queue_name, exchange=exchange, routing_key=routing_key)
    channel.basic_consume(queue=queue_name, on_message_callback=callback, auto_ack=auto_ack)
    channel.start_consuming()


class Channel:
    def __init__(self, schema: Schema):
        self.schema = schema

    def connect(self):
        # FIXME implement this
        pass

    def close(self):
        # FIXME implement this
        pass

    def send(self, event):
        # I don't know how to type Event there for Python but it would be
        # nice if we could, somehow.
        rendered = self.schema.dump(event)
        # FIXME: transmit rendered over the underlying channel
        pass

    def listen(self, callback: Callable[[Any], None], pattern: str = '#'):
        def unwrapping_callback(message):
            event = self.schema.load(message)
            callback(event)
        # FIXME: ask the underlying channel to call us back
        # but only for messages matching pattern
        pass

    def __enter__(self):
        self.connect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


# Predefined event channels for ease of use
workflow_events = Channel(WorkflowEventSchema())
